/*==============================================================================


   Purpose:    determine MHz rate the processor is running


   Author:     Rudolf Berrendorf
               Forschungszentrum Juelich GmbH
               Zentralinstitut fuer Angewandte Mathematik
               D-52425 Juelich
               Tel: ++49 2461 61-3218
               r.berrendorf@fz-juelich.de


   Date:       02.07.1999


   SUPPORT AND DISCLAIMER

   The software in this file (the "Software") is an unsupported product and is
   provided by Zentralinstitut fuer Angewandte Mathematik, Forschungszentrum
   Juelich GmbH (ZAM) "AS IS". We welcome comments, suggestions, and problem
   reports and ask that you mail those to kfazam@kfa-juelich.de; however, we
   make no promises to respond to any item.

   ZAM makes no warrranties or representations, expressed or implied, including,
   but not limited to, implied warranties of merchantability and fitness for a
   particular purpose, as to any element of the Software or any support
   provided in connection with the Software. In no event shall ZAM be
   responsible for any damages, including but not limited to consequential
   damages, arising from or relating to any use of the Software or related
   support.


   COPYRIGHT

   (c) COPYRIGHT 1999
   Zentralinstitut fuer Angewandte Mathematik, Forschungszentrum Juelich GmbH

   The Software is copyrighted. For research purposes, you can use it for free.
   If you plan to use it commercially, please contact zam@fz-juelich.de.


==============================================================================*/

#include "determine_mhz.h"


#if defined(ALPHA_CRAY)
#include <unistd.h>

#elif defined(ALPHA_DEC)
#include <sys/sysinfo.h>
#include <machine/hal_sysinfo.h>

#elif defined(MIPS)
#include <stdio.h>

#elif defined(SPARC)
#include <stdio.h>
#include <stdlib.h>

#elif defined(PENTIUM_LINUX)
#include <stdio.h>
#include <string.h>

#endif


/*                                                                            */
/*============================================================================*/
/* returns MHz-rate of the processor (or -1 if not available)                 */


int
PCL_determine_mhz_rate(void)

{
  /*--------------------------------------------------------------------------*/

#if defined(ALPHA_CRAY)
  {
    long cycle_time;

    /* get cycle time in picoseconds */
    if((cycle_time = sysconf(_SC_CRAY_CPCYCLE)) == -1)
      return -1;
    else
      return (long)(1e6 / cycle_time);
  }

  /*--------------------------------------------------------------------------*/

#elif defined(ALPHA_DEC)
  {
    struct cpu_info cpu_info;

    if(getsysinfo(GSI_CPU_INFO, &cpu_info, sizeof(cpu_info), NULL, NULL, NULL)
       < 1)
      return -1;
    else
      return cpu_info.mhz;
  }

  /*--------------------------------------------------------------------------*/

#elif defined(MIPS)

  {
    int dummy, mhz, res;
    FILE *f;
    char cmd_line[80], fname[L_tmpnam];

    /*??? system call takes very long */
    /* get system configuration and put output into file */
    sprintf(cmd_line, "/sbin/hinv -c processor >%s", tmpnam(fname));
    if(system(cmd_line) == -1)
      {
	remove(fname);
	return -1;
      }

    /* open output file */
    if((f = fopen(fname, "r")) == NULL)
      {
	remove(fname);
	return -1;
      }

    /* get second item on first line */
    res = fscanf(f, "%d %d", &dummy, &mhz);

    /* remove temporary file */
    remove(fname);

    if(res != 2)
      return -1;
    else
      return mhz;
  }

  /*--------------------------------------------------------------------------*/

#elif defined(SPARC)

  {
    int mhz;
    char line[256], cmd[80];
    FILE *f;
    char cmd_line[80], fname[L_tmpnam];

    /*??? system call takes very long */
    /* get system configuration and put output into file */
    sprintf(cmd_line, "/usr/sbin/prtconf -vp >%s", tmpnam(fname));
    if(system(cmd_line) == -1)
      {
	remove(fname);
	return -1;
      }

    /* open output file */
    if((f = fopen(fname, "r")) == NULL)
      {
	remove(fname);
	return -1;
      }

    /* ignore all lines until we reach something with a sparc line */
    while(fgets(line, 256, f) != NULL)
      {
	if((sscanf(line, "%s", cmd) == 1)
	   && !strcmp(cmd, "sparc-version:"))
	  break;
      }

    /* then read until we find clock frequency */
    while(fgets(line, 256, f) != NULL)
      {
	if((sscanf(line, "%s %x", cmd, &mhz) == 2)
	   && !strcmp(cmd, "clock-frequency:"))
	  break;
      }

    /* remove temporary file */
    remove(fname);

    /* if everything wqent ok, return mhz */
    if(strcmp(cmd, "clock-frequency:"))
      return -1;
    else
      return mhz / 1000000;
  }

  /*--------------------------------------------------------------------------*/

#elif defined(POWERPC)

  /*??? not yet done */
  return -1;

  /*--------------------------------------------------------------------------*/

#elif defined(PENTIUM_LINUX)
  {
    int res, mhz;
    char line[256], *s, search_str[] = "cpu MHz";
    FILE *f;

    /* open proc/cpuinfo */
    if((f = fopen("/proc/cpuinfo", "r")) == NULL)
      return -1;

    /* ignore all lines until we reach MHz information */
    res = -1;
    while(fgets(line, 256, f) != NULL)
      {
	if(strstr(line, search_str) != NULL)
	  {
	    /* ignore all characters in line up to : */
	    for(s = line; *s && (*s != ':'); ++s)
	      ;
	    /* get MHz number */
	    if(*s && (sscanf(s+1, "%d", &mhz) == 1))
	      {
		res = mhz;
		break;
	      }
	  }
      }

    return res;
  } 

  /*--------------------------------------------------------------------------*/

#elif defined(PENTIUM_NT)

  /* not implemented yet */
  return -1;

  /*--------------------------------------------------------------------------*/

#else
#error "unsupported system"
#endif
}


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/
