#pragma ident   "@(#)perfmon.c 1.3     97/04/30 SMI"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/modctl.h>
#include <sys/x_call.h>
#include "perfmon.h"

static int perfmon_identify(dev_info_t *);
static int perfmon_attach(dev_info_t *, ddi_attach_cmd_t);
static int perfmon_detach(dev_info_t *, ddi_detach_cmd_t);
static int perfmon_info(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
			void **result);
static int perfmon_open(dev_t *, int, int, cred_t *);
static int perfmon_close(dev_t, int, int, cred_t *);
static int perfmon_ioctl(dev_t, int, int, int, cred_t *, int *);
static int perfmon_init(void);
static int perfmon_cleanup(void);

static dev_info_t *perfmon_devi;

static kmutex_t	perfmon_lock;

/*
 * Define device operations.  These are the functions that are called
 * from user-land by open/read/write/ioctl etc.
 */
static struct cb_ops	perfmon_cb_ops = {
    perfmon_open,	/* open */
    perfmon_close,	/* close */
    nodev,		/* strategy */
    nodev,		/* print */
    nodev,		/* dump */
    nodev,		/* read */
    nodev,		/* write */
    perfmon_ioctl,	/* ioctl */
    nodev,		/* devmap */
    nodev,		/* mmap */
    nodev,		/* segmap */
    nochpoll,		/* poll */
    ddi_prop_op,	/* prop_op */
    0,			/* streamtab  */
    D_NEW | D_MP	/* Driver compatibility flag */
};

/*
 * Define device driver entry points.  These functions are called by
 * the kernel to have us do housekeeping, etc.
 */
static struct dev_ops	perfmon_ops = {
    DEVO_REV,		/* devo_rev, */
    0,			/* refcnt  */
    perfmon_info,	/* get_dev_info */
    perfmon_identify,	/* identify */
    nulldev,		/* probe */
    perfmon_attach,	/* attach */
    perfmon_detach,	/* detach */
    nodev,		/* reset */
    &perfmon_cb_ops,	/* driver operations */
    (struct bus_ops *)0	/* no bus operations */
};

/* Define some misc structures that the kernel needs to find in our module */
static struct modldrv modldrv = {
    &mod_driverops,
    "Performance Monitor Driver",
    &perfmon_ops,
};

static struct modlinkage modlinkage = {
    MODREV_1,
    &modldrv,
    NULL
};

/*
 * This is the driver entry point that gets called right after we've
 * been loaded into kernel space.
 */
int _init(void)
{
    int rc;
    int i;

    /* Initialize our mutex */
    mutex_init(&perfmon_lock, "perfmon driver lock",
	MUTEX_DEFAULT, (void *)0);

    if ((rc = mod_install(&modlinkage)) != 0)  {
	mutex_destroy(&perfmon_lock);
	return (rc);
    }
    return (0);
}

/*
 * This is the driver entry point that gets called right before we're
 * unloaded.  We need to do our housecleaning here.
 */
int _fini(void)
{
    int rc;

    if ((rc = mod_remove(&modlinkage)) == 0) {
	(void) perfmon_cleanup();
	mutex_destroy(&perfmon_lock);
	return (0);
    }
    return (rc);
}

/* The kernel is poking us.  Give it what it wants. */
int _info(struct modinfo *modinfop)
{
    return (mod_info(&modlinkage, modinfop));
}

/* Identify ourselves. */
static int perfmon_identify(dev_info_t *devi)
{
    /*
     * If we've finished initialization and our device info structure is
     * in place, we're OK.  Otherwise, we're not ready yet.
     */
    if (strcmp(ddi_get_name(devi), "perfmon") == 0) {
	return (DDI_IDENTIFIED);
    } else {
	return (DDI_NOT_IDENTIFIED);
    }
}

/*
 * We've just been loaded and need to create our device node and do
 * some initialization stuff
 */
static int perfmon_attach(dev_info_t *devi, ddi_attach_cmd_t cmd)
{
    switch (cmd) {
	case DDI_ATTACH:
	    /* Create our device node */
	    if (ddi_create_minor_node(devi, "perfmon", S_IFCHR, 0, NULL, NULL)
		    == DDI_FAILURE) {
		ddi_remove_minor_node(devi, NULL);
		return (DDI_FAILURE);
	    }
	    perfmon_devi = devi;

	    if (perfmon_init()) {
		ddi_remove_minor_node(devi, NULL);
		return(DDI_FAILURE);
	    }

	    return (DDI_SUCCESS);
	    break;

	case DDI_RESUME:
	    return (DDI_SUCCESS);

	default:
	    return (DDI_FAILURE);
    }
}

/*
 * We're being unloaded, so get rid of our device node and clean house.
 */
static int perfmon_detach(dev_info_t *devi, ddi_detach_cmd_t cmd)
{
    switch (cmd) {
	case DDI_DETACH:
	    ddi_remove_minor_node(devi, NULL);
	    return (DDI_SUCCESS);
	    break;

	case DDI_SUSPEND:
	    return (DDI_SUCCESS);

	default:
	    return (DDI_FAILURE);
    }
}

/*
 * We need to return information about ourselves.
 */
static int perfmon_info(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
    void **result)
{
    register int error;

    switch (infocmd) {
	case DDI_INFO_DEVT2DEVINFO:
	    *result = (void *) perfmon_devi;
	    error = DDI_SUCCESS;
	    break;
	case DDI_INFO_DEVT2INSTANCE:
	    *result = (void *)0;
	    error = DDI_SUCCESS;
	    break;
	default:
	    error = DDI_FAILURE;
    }
    return (error);
}

/*
 * Do all of our general initialization tasks.  If we need to worry
 * about unwinding in the event of an error, we should do that back
 * up in _init().
 */
static int perfmon_init(void)
{
#ifdef PERFMON_DEBUG
    cmn_err(CE_NOTE, "perfmon: init() called\n");
#endif

    /*
     * Do a cross-call to all CPUs to enable user-level TICK
     * reads.  This routine must be run at a high PIL so that
     * a TICK_CMPR or something doesn't go off in the middle
     * of things.  Lucky for us, cross calls on sun4u run at
     * PIL == 13, which is good enough to keep us out of
     * trouble.  We also want to get everyone's attention so
     * that when we return, we are sure that our routine has
     * been completed.
     */
    xc_attention(cpu_ready_set);
    xc_all((u_int (*)())pm_set_rdtick_mode, 0, 0);
    xc_dismissed(cpu_ready_set);

    return (0);
}

/*
 * Handle ioctl() requests from user-land.
 */
static int perfmon_ioctl(dev_t dev, int cmd, int arg, int mode,
    cred_t *cred_p, int *rval_p)
{
    static void (*cache_flush)(void) = NULL;
    int try_lookup = 1;
    unsigned long long tmp;

#ifdef PERFMON_DEBUG
    cmn_err(CE_NOTE, "perfmon: ioctl() called\n");
#endif

    switch (cmd) {
	/* We get a longlong from the user and stuff it in the PCR */
	case PERFMON_SETPCR:
	    if (copyin((void *)arg, (void *)&tmp, sizeof(tmp)))
		return (EFAULT);
#ifdef PERFMON_DEBUG
	    cmn_err(CE_NOTE, "perfmon: write got tmp = %llx\n", tmp);
#endif
	    pm_set_pcr(tmp);
	    /* xc_all((u_int (*)())pm_set_pcr, tmp, 0); */
	    break;

	/* Get the PCR and return it to the user's buffer */
	case PERFMON_GETPCR:
	    tmp = pm_get_pcr();
	    cmn_err(CE_NOTE, "perfmon: read got tmp = %llx\n", tmp);
	    if (copyout((void *)&tmp, (void *)arg, sizeof(tmp)))
		return (EFAULT);
	    break;

	/* Flush the CPU's cache */
	case PERFMON_FLUSH_CACHE:

	/* Flush every CPU's cache */
	case PERFMON_FLUSH_ALL_CACHES:

	    /* If we haven't looked up cpu_flush_ecache yet, do it now */
	    if (cache_flush == NULL && try_lookup) {
		/*
		 * Just to be MP-safe, grab a lock before we go messing
		 * around.
		 */
		mutex_enter(&perfmon_lock);

		/* We don't need to do the lookup more than once */
		if (cache_flush == NULL && try_lookup) {
		    cache_flush =
			(void (*)(void))kobj_getsymvalue("cpu_flush_ecache", 0);
		    try_lookup = 0;
		}

		if (cache_flush == NULL)
		    cmn_err(CE_NOTE, "perfmon: "
			"CPU cache flushing not supported on this machine");
		else
		    cmn_err(CE_NOTE, "perfmon: "
			"CPU cache flushing is supported");

		mutex_exit(&perfmon_lock);
	    }

	    /*
	     * If we can't find the function we need, then this operation
	     * isn't supported.
	     */
	    if (cache_flush == NULL)
		return(ENOTSUP);
	    else
		/* if (cmd == PERFMON_FLUSH_ALL_CACHES) */
		    /* xc_all((u_int (*)())cache_flush, 0, 0); */
		/* else */
		    (cache_flush)();
	    break;

	default:
	    /* invalid request */
	    *rval_p = -1;
	    return (EINVAL);
    }

    return (0);
}

/*
 * Someone is trying to open() us.  Go ahead and let them.
 */
static int perfmon_open(dev_t *devp, int flag, int otyp, struct cred *cred)
{
#ifdef	PERFMON_DEBUG
    cmn_err(CE_NOTE, "perfmon_open: dev: %x\n", getminor(*devp));
#endif	/* PERFMON_DEBUG */

    return (0);
}

/*
 * Someone is closing us.  Nothing to worry about here.
 */
static int perfmon_close(dev_t dev, int flag, int otyp, struct cred *cred)
{
#ifdef	PERFMON_DEBUG
    cmn_err(CE_NOTE, "perfmon: dev: %x\n", getminor(dev));
#endif	/* PERFMON_DEBUG */

    return (0);
}

/*
 * Take care of any housecleaning that needs to be done on our way out.
 */
static int perfmon_cleanup(void)
{
#ifdef PERFMON_DEBUG
    cmn_err(CE_NOTE, "perfmon: cleanup() called\n");
#endif

    /* Should we reset the priv bit in the TICK register? */
    return (0);
}
